// "WinCrt Unit", Copyright (c) 2003-2010 by Stefan Berinde
//   Version: as WinGraph unit
//
// Source code: Free Pascal 2.4 (http://www.freepascal.org) &
//              Delphi 7
//
// Unit dependences:
//   windows  (standard)
//   messages (standard, only for Delphi)
//   wingraph
//
{$MODE DELPHI}

Unit wincrt;

Interface

{caret format: block or underline}

Var CaretBlock: boolean = true;

{caret blink rate in tenths of a second}

Var BlinkRate: word = 2;

{keyboard exported routines}
Procedure Delay(ms:word);
Function KeyPressed: boolean;
Procedure ReadBuf(out buf:shortstring; maxchar:byte);
Function ReadKey: char;
Procedure Sound(hz,dur:word);
Procedure WriteBuf(buf:shortstring);


Implementation

Uses windows{$IFNDEF FPC},messages{$ENDIF},wingraph;

Const KeyBufSize = 55;

Var protect_keyboard        : TRTLCriticalSection;
  nr_readkey,nr_inputkey  : longint;
  keyBuf                  : array[1..KeyBufSize] Of char;
  old_TextSettings        : TextSettingsType;
  textX,textY,textW,textH,
  maxX,maxY               : smallint;

{internal routines}

Procedure InitKeyBuf;
Begin
  nr_readkey := 1;
  nr_inputkey := 1;
End;

Procedure IncKeyCyclic(Var nr: longint);
Begin
  Inc(nr);
  If (nr > KeyBufSize) Then nr := 1;
End;

Procedure AddKey(c:char);
Begin
  EnterCriticalSection(protect_keyboard);
  KeyBuf[nr_inputkey] := c;
  IncKeyCyclic(nr_inputkey);
  If (nr_readkey = nr_inputkey) Then
    Begin
      If (KeyBuf[nr_readkey] = #0) Then IncKeyCyclic(nr_readkey);
      IncKeyCyclic(nr_readkey);
    End;
  LeaveCriticalSection(protect_keyboard);
End;

Procedure AddExtKey(c:char);
Begin
  AddKey(#0);
  AddKey(c);
End;

Procedure TranslateKeys(code:WPARAM);

Var shift_key,ctrl_key,alt_key: boolean;
Begin
  shift_key := (GetKeyState(VK_SHIFT) < 0);
  ctrl_key := (GetKeyState(VK_CONTROL) < 0);
  alt_key := (GetKeyState(VK_MENU) < 0);
  Case code Of
    VK_SPACE: If alt_key Then AddExtKey(#11);
    VK_TAB: If ctrl_key Then AddKey(#30);
    VK_BACK: If alt_key Then AddExtKey(#14);
    VK_RETURN: If alt_key Then AddExtKey(#166);
    VK_APPS: AddExtKey(#151);
    VK_INSERT: If ctrl_key Then AddExtKey(#146)
               Else
                 If alt_key Then AddExtKey(#162)
               Else AddExtKey(#82);
    VK_DELETE: If ctrl_key Then AddExtKey(#147)
               Else
                 If alt_key Then AddExtKey(#163)
               Else AddExtKey(#83);
    VK_HOME: If ctrl_key Then AddExtKey(#119)
             Else
               If alt_key Then AddExtKey(#164)
             Else AddExtKey(#71);
    VK_END: If ctrl_key Then AddExtKey(#117)
            Else
              If alt_key Then AddExtKey(#165)
            Else AddExtKey(#79);
    VK_NEXT: If ctrl_key Then AddExtKey(#118)
             Else
               If alt_key Then AddExtKey(#161)
             Else AddExtKey(#81);
    VK_PRIOR: If ctrl_key Then AddExtKey(#132)
              Else
                If alt_key Then AddExtKey(#153)
              Else AddExtKey(#73);
    VK_UP: If ctrl_key Then AddExtKey(#141)
           Else
             If alt_key Then AddExtKey(#152)
           Else AddExtKey(#72);
    VK_DOWN: If ctrl_key Then AddExtKey(#145)
             Else
               If alt_key Then AddExtKey(#160)
             Else AddExtKey(#80);
    VK_LEFT: If ctrl_key Then AddExtKey(#115)
             Else
               If alt_key Then AddExtKey(#155)
             Else AddExtKey(#75);
    VK_RIGHT: If ctrl_key Then AddExtKey(#116)
              Else
                If alt_key Then AddExtKey(#157)
              Else AddExtKey(#77);
    VK_F1..VK_F10: If shift_key Then AddExtKey(chr(code-28))
                   Else
                     If ctrl_key Then AddExtKey(chr(code-18))
                   Else
                     If alt_key Then AddExtKey(chr(code-8))
                   Else AddExtKey(chr(code-53));
    VK_F11,VK_F12: If shift_key Then AddExtKey(chr(code+13))
                   Else
                     If ctrl_key Then AddExtKey(chr(code+15))
                   Else
                     If alt_key Then AddExtKey(chr(code+17))
                   Else AddExtKey(chr(code+11));
    VK_PAUSE: If alt_key Then AddExtKey(#169)
              Else
                If Not(ctrl_key) Then AddExtKey(#12);
    VK_CLEAR: If ctrl_key Then AddExtKey(#143)
              Else AddExtKey(#76);
    //this is numpad 5 + numlock off
    VK_DIVIDE: If ctrl_key Then AddExtKey(#148)
               Else
                 If alt_key Then AddExtKey(#69);
    VK_MULTIPLY: If ctrl_key Then AddExtKey(#149)
                 Else
                   If alt_key Then AddExtKey(#70);
    VK_SUBTRACT: If ctrl_key Then AddExtKey(#142)
                 Else
                   If alt_key Then AddExtKey(#74);
    VK_ADD: If ctrl_key Then AddExtKey(#144)
            Else
              If alt_key Then AddExtKey(#78);
    VK_DECIMAL: If ctrl_key Then AddExtKey(#150)
                Else
                  If alt_key Then AddExtKey(#114);
    Else
      If ctrl_key Then Case code Of
                         ord('0')          : AddExtKey(#10);
                         ord('1')..ord('9'): AddExtKey(chr(code-48));
        End;
    If alt_key Then Case code Of
                      ord('A'): AddExtKey(#30);
                      ord('B'): AddExtKey(#48);
                      ord('C'): AddExtKey(#46);
                      ord('D'): AddExtKey(#32);
                      ord('E'): AddExtKey(#18);
                      ord('F'): AddExtKey(#33);
                      ord('G'): AddExtKey(#34);
                      ord('H'): AddExtKey(#35);
                      ord('I'): AddExtKey(#23);
                      ord('J'): AddExtKey(#36);
                      ord('K'): AddExtKey(#37);
                      ord('L'): AddExtKey(#38);
                      ord('M'): AddExtKey(#50);
                      ord('N'): AddExtKey(#49);
                      ord('O'): AddExtKey(#24);
                      ord('P'): AddExtKey(#25);
                      ord('Q'): AddExtKey(#16);
                      ord('R'): AddExtKey(#19);
                      ord('S'): AddExtKey(#31);
                      ord('T'): AddExtKey(#20);
                      ord('U'): AddExtKey(#22);
                      ord('V'): AddExtKey(#47);
                      ord('W'): AddExtKey(#17);
                      ord('X'): AddExtKey(#45);
                      ord('Y'): AddExtKey(#21);
                      ord('Z'): AddExtKey(#44);
                      ord('0')          : AddExtKey(#129);
                      ord('1')..ord('9'): AddExtKey(chr(code+71));
      End;
  End;
End;

Function WinCrtProc(grHandle:HWND; mess:UINT; wParam:WPARAM;
                    lParam:LPARAM): LRESULT;
stdcall;
Begin
  Result := 0;
  Case mess Of
    WM_CREATE:
               Begin
                 InitializeCriticalSection(protect_keyboard);
                 InitKeyBuf;
               End;
    WM_CHAR: AddKey(chr(wparam));
    WM_KEYDOWN: If (wParam <> VK_SHIFT) And (wParam <> VK_CONTROL) Then TranslateKeys(wParam);
    WM_SYSKEYDOWN: If (wParam <> VK_MENU) Then TranslateKeys(wParam);
    WM_CLOSE: AddExtKey(#107);
    WM_DESTROY: DeleteCriticalSection(protect_keyboard);
  End;
End;

Procedure CheckNewLine;

Var size  : longword;
  screen: pointer;
Begin
  If (textX+textW > maxX) Then
    Begin
      textX := 0;
      Inc(textY,textH);
    End;
  If (textY+textH > maxY) Then //scroll entire text upwards
    Begin
      Repeat
        Dec(textY,textH);
      Until (textY+textH <= maxY);
      size := ImageSize(0,textH,maxX,maxY);
      GetMem(screen,size);
      GetImage(0,textH,maxX,maxY,screen^);
      ClearViewPort;
      PutImage(0,0,screen^,CopyPut);
      FreeMem(screen);
    End;
End;

Procedure DrawCaret(nr:longint);
Begin

  Case nr Of
    0: SetFillStyle(SolidFill,GetBkColor);
    1: SetFillStyle(SolidFill,GetColor);
  End;
  If CaretBlock Or (nr = 0) Then Bar(textX,textY+textH-1,textX+textW,textY+textH)
  //textX+textW,textY+textH)   x1:=textX; y1:=textY+textH-1; x2:=textX+textW; y2:=textY+textH;
  Else Bar(textX,textY+textH-1,textX+textW,textY+textH);
End;

Procedure TextSettings;

Var viewport: ViewPortType;
Begin
  With old_TextSettings Do
    Begin
      SetTextStyle(DefaultFont Or (font Div $10) shl 4,0,charsize);
      //keep font format
      SetTextJustify(LeftText,TopText);
    End;
  textX := GetX;
  textY := GetY;
  textW := TextWidth('W');
  textH := TextHeight('H');
  GetViewSettings(viewport);
  With viewport Do
    Begin
      maxX := x2-x1;
      maxY := y2-y1;
    End;
End;

{keyboard routines}

Procedure Delay(ms:word);
Begin
  Sleep(ms);
End;

Function KeyPressed: boolean;
Begin
  Result := (nr_readkey <> nr_inputkey);
End;

Procedure ReadBuf(out buf:shortstring; maxchar:byte);

Var old_FillSettings     : FillSettingsType;
  nrpass,nrchar,nrcaret: longint;
  ch                   : char;
Begin
  If GraphEnabled Then
    Begin
      GetTextSettings(old_TextSettings);
      GetFillSettings(old_FillSettings);
      TextSettings;
      CheckNewLine;
      nrpass := 0;
      nrcaret := 0;
      nrchar := 0;
      ch := #0;
      If (maxchar <= 0) Then maxchar := 255;
      Repeat
        If (nrpass = 0) Then
          Begin
            nrcaret := 1-nrcaret;
            DrawCaret(nrcaret);
            nrpass := 10*BlinkRate;
          End
        Else Dec(nrpass);
        If KeyPressed Then
          Begin
            ch := ReadKey;
            Case ch Of
              #32..#126:
                         Begin
                           If (nrcaret = 1) Then
                             Begin
                               DrawCaret(0);
                               nrcaret := 0;
                             End;
                           Inc(nrchar);
                           buf[nrchar] := ch;
                           OutTextXY(textX,textY,ch);
                           Inc(textX,textW);
                           CheckNewLine;
                           nrpass := 0;
                         End;
              #8: If (nrchar > 0) And (textX > 0) Then
                    Begin
                      If (nrcaret = 1) Then
                        Begin
                          DrawCaret(0);
                          nrcaret := 0;
                        End;
                      Dec(nrchar);
                      Dec(textX,textW);
                      nrpass := 0;
                    End;
              #0: ReadKey;
            End;
          End;
        Sleep(10);
      Until (ch = #13) Or (nrchar = maxchar) Or CloseGraphRequest;
      If (nrcaret = 1) Then DrawCaret(0);
      buf[0] := Chr(nrchar);
      MoveTo(0,textY+textH);
      With old_TextSettings Do
        Begin
          SetTextStyle(font,direction,charsize);
          SetTextJustify(horiz,vert);
        End;
      With old_FillSettings Do
        SetFillStyle(pattern,color);
    End;
End;

Function ReadKey: char;
Begin
  If GraphEnabled Then
    Begin
      While (nr_readkey = nr_inputkey) Do
        Sleep(10);
      EnterCriticalSection(protect_keyboard);
      Result := KeyBuf[nr_readkey];
      IncKeyCyclic(nr_readkey);
      LeaveCriticalSection(protect_keyboard);
    End
  Else Result := #0;
End;

Procedure Sound(hz,dur:word);
Begin
  Beep(hz,dur);
End;

Procedure WriteBuf(buf:shortstring);

Var old_FillSettings: FillSettingsType;
  nrchar          : longint;
  ch              : char;
Begin
  If GraphEnabled Then
    Begin
      GetTextSettings(old_TextSettings);
      GetFillSettings(old_FillSettings);
      TextSettings;
      CheckNewLine;
      For nrchar:=1 To Length(buf) Do
        Begin
          ch := buf[nrchar];
          Case ch Of
            #32..#126:
                       Begin
                         DrawCaret(0);
                         OutTextXY(textX,textY,ch);
                         Inc(textX,textW);
                         CheckNewLine;
                       End;
            #13:
                 Begin
                   textX := maxX;
                   CheckNewLine;
                 End;
          End;
        End;
      MoveTo(textX,textY);
      With old_TextSettings Do
        Begin
          SetTextStyle(font,direction,charsize);
          SetTextJustify(horiz,vert);
        End;
      With old_FillSettings Do
        SetFillStyle(pattern,color);
    End;
End;

initialization
KeyboardHook := @WinCrtProc;
End.
